/*
 * I am the beam position adjuster engine.
 *
 * Macro parameters:
 *   P  PV prefix (e.g. 17idb:)
 *   A  Beam position adjuster name (e.g. Bpa3)
 */

program beampad
option +r;


%%#include <math.h>

%%#include <epicsAssert.h>
%%#include <errlog.h>


#define PROGRAM_NAME "beampad"

#define OFFSET_TABLE_SIZE 20
#define OFFSET_TABLE_NIL_LIFE_NUM -1
#define ENERGY_CHANGED_RATE_LIMITING_DELAY 0.25


double OffsetX;
  assign OffsetX to "{P}{A}:OffsetX";
long IndexOfOffsetX;

double OffsetY;
  assign OffsetY to "{P}{A}:OffsetY";
long IndexOfOffsetY;

double Energy;
  assign Energy to "{P}{A}:Energy";
  monitor Energy;
  evflag EnergyChanged;
  sync Energy EnergyChanged;

double EnergyBand;
  assign EnergyBand to "{P}{A}:EnergyBand";
  monitor EnergyBand;
  evflag EnergyBandChanged;
  sync EnergyBand EnergyBandChanged;

int OffsetTableLifeNum[OFFSET_TABLE_SIZE];
  assign OffsetTableLifeNum[0]  to "{P}{A}:LifeNum1";
  assign OffsetTableLifeNum[1]  to "{P}{A}:LifeNum2";
  assign OffsetTableLifeNum[2]  to "{P}{A}:LifeNum3";
  assign OffsetTableLifeNum[3]  to "{P}{A}:LifeNum4";
  assign OffsetTableLifeNum[4]  to "{P}{A}:LifeNum5";
  assign OffsetTableLifeNum[5]  to "{P}{A}:LifeNum6";
  assign OffsetTableLifeNum[6]  to "{P}{A}:LifeNum7";
  assign OffsetTableLifeNum[7]  to "{P}{A}:LifeNum8";
  assign OffsetTableLifeNum[8]  to "{P}{A}:LifeNum9";
  assign OffsetTableLifeNum[9]  to "{P}{A}:LifeNum10";
  assign OffsetTableLifeNum[10] to "{P}{A}:LifeNum11";
  assign OffsetTableLifeNum[11] to "{P}{A}:LifeNum12";
  assign OffsetTableLifeNum[12] to "{P}{A}:LifeNum13";
  assign OffsetTableLifeNum[13] to "{P}{A}:LifeNum14";
  assign OffsetTableLifeNum[14] to "{P}{A}:LifeNum15";
  assign OffsetTableLifeNum[15] to "{P}{A}:LifeNum16";
  assign OffsetTableLifeNum[16] to "{P}{A}:LifeNum17";
  assign OffsetTableLifeNum[17] to "{P}{A}:LifeNum18";
  assign OffsetTableLifeNum[18] to "{P}{A}:LifeNum19";
  assign OffsetTableLifeNum[19] to "{P}{A}:LifeNum20";
long IndexOfOffsetTableLifeNum[OFFSET_TABLE_SIZE];

short OffsetTableSelected[OFFSET_TABLE_SIZE];
  assign OffsetTableSelected[0]  to "{P}{A}:Selected1";
  assign OffsetTableSelected[1]  to "{P}{A}:Selected2";
  assign OffsetTableSelected[2]  to "{P}{A}:Selected3";
  assign OffsetTableSelected[3]  to "{P}{A}:Selected4";
  assign OffsetTableSelected[4]  to "{P}{A}:Selected5";
  assign OffsetTableSelected[5]  to "{P}{A}:Selected6";
  assign OffsetTableSelected[6]  to "{P}{A}:Selected7";
  assign OffsetTableSelected[7]  to "{P}{A}:Selected8";
  assign OffsetTableSelected[8]  to "{P}{A}:Selected9";
  assign OffsetTableSelected[9]  to "{P}{A}:Selected10";
  assign OffsetTableSelected[10] to "{P}{A}:Selected11";
  assign OffsetTableSelected[11] to "{P}{A}:Selected12";
  assign OffsetTableSelected[12] to "{P}{A}:Selected13";
  assign OffsetTableSelected[13] to "{P}{A}:Selected14";
  assign OffsetTableSelected[14] to "{P}{A}:Selected15";
  assign OffsetTableSelected[15] to "{P}{A}:Selected16";
  assign OffsetTableSelected[16] to "{P}{A}:Selected17";
  assign OffsetTableSelected[17] to "{P}{A}:Selected18";
  assign OffsetTableSelected[18] to "{P}{A}:Selected19";
  assign OffsetTableSelected[19] to "{P}{A}:Selected20";
long IndexOfOffsetTableSelected[OFFSET_TABLE_SIZE];

double OffsetTableEnergy[OFFSET_TABLE_SIZE];
  assign OffsetTableEnergy[0]  to "{P}{A}:Energy1";
  assign OffsetTableEnergy[1]  to "{P}{A}:Energy2";
  assign OffsetTableEnergy[2]  to "{P}{A}:Energy3";
  assign OffsetTableEnergy[3]  to "{P}{A}:Energy4";
  assign OffsetTableEnergy[4]  to "{P}{A}:Energy5";
  assign OffsetTableEnergy[5]  to "{P}{A}:Energy6";
  assign OffsetTableEnergy[6]  to "{P}{A}:Energy7";
  assign OffsetTableEnergy[7]  to "{P}{A}:Energy8";
  assign OffsetTableEnergy[8]  to "{P}{A}:Energy9";
  assign OffsetTableEnergy[9]  to "{P}{A}:Energy10";
  assign OffsetTableEnergy[10] to "{P}{A}:Energy11";
  assign OffsetTableEnergy[11] to "{P}{A}:Energy12";
  assign OffsetTableEnergy[12] to "{P}{A}:Energy13";
  assign OffsetTableEnergy[13] to "{P}{A}:Energy14";
  assign OffsetTableEnergy[14] to "{P}{A}:Energy15";
  assign OffsetTableEnergy[15] to "{P}{A}:Energy16";
  assign OffsetTableEnergy[16] to "{P}{A}:Energy17";
  assign OffsetTableEnergy[17] to "{P}{A}:Energy18";
  assign OffsetTableEnergy[18] to "{P}{A}:Energy19";
  assign OffsetTableEnergy[19] to "{P}{A}:Energy20";
long IndexOfOffsetTableEnergy[OFFSET_TABLE_SIZE];

double OffsetTableX[OFFSET_TABLE_SIZE];
  assign OffsetTableX[0]  to "{P}{A}:OffsetX1";
  assign OffsetTableX[1]  to "{P}{A}:OffsetX2";
  assign OffsetTableX[2]  to "{P}{A}:OffsetX3";
  assign OffsetTableX[3]  to "{P}{A}:OffsetX4";
  assign OffsetTableX[4]  to "{P}{A}:OffsetX5";
  assign OffsetTableX[5]  to "{P}{A}:OffsetX6";
  assign OffsetTableX[6]  to "{P}{A}:OffsetX7";
  assign OffsetTableX[7]  to "{P}{A}:OffsetX8";
  assign OffsetTableX[8]  to "{P}{A}:OffsetX9";
  assign OffsetTableX[9]  to "{P}{A}:OffsetX10";
  assign OffsetTableX[10] to "{P}{A}:OffsetX11";
  assign OffsetTableX[11] to "{P}{A}:OffsetX12";
  assign OffsetTableX[12] to "{P}{A}:OffsetX13";
  assign OffsetTableX[13] to "{P}{A}:OffsetX14";
  assign OffsetTableX[14] to "{P}{A}:OffsetX15";
  assign OffsetTableX[15] to "{P}{A}:OffsetX16";
  assign OffsetTableX[16] to "{P}{A}:OffsetX17";
  assign OffsetTableX[17] to "{P}{A}:OffsetX18";
  assign OffsetTableX[18] to "{P}{A}:OffsetX19";
  assign OffsetTableX[19] to "{P}{A}:OffsetX20";
  monitor OffsetTableX;
  evflag OffsetTableXChanged;
  sync OffsetTableX OffsetTableXChanged;
long IndexOfOffsetTableX[OFFSET_TABLE_SIZE];

double OffsetTableY[OFFSET_TABLE_SIZE];
  assign OffsetTableY[0]  to "{P}{A}:OffsetY1";
  assign OffsetTableY[1]  to "{P}{A}:OffsetY2";
  assign OffsetTableY[2]  to "{P}{A}:OffsetY3";
  assign OffsetTableY[3]  to "{P}{A}:OffsetY4";
  assign OffsetTableY[4]  to "{P}{A}:OffsetY5";
  assign OffsetTableY[5]  to "{P}{A}:OffsetY6";
  assign OffsetTableY[6]  to "{P}{A}:OffsetY7";
  assign OffsetTableY[7]  to "{P}{A}:OffsetY8";
  assign OffsetTableY[8]  to "{P}{A}:OffsetY9";
  assign OffsetTableY[9]  to "{P}{A}:OffsetY10";
  assign OffsetTableY[10] to "{P}{A}:OffsetY11";
  assign OffsetTableY[11] to "{P}{A}:OffsetY12";
  assign OffsetTableY[12] to "{P}{A}:OffsetY13";
  assign OffsetTableY[13] to "{P}{A}:OffsetY14";
  assign OffsetTableY[14] to "{P}{A}:OffsetY15";
  assign OffsetTableY[15] to "{P}{A}:OffsetY16";
  assign OffsetTableY[16] to "{P}{A}:OffsetY17";
  assign OffsetTableY[17] to "{P}{A}:OffsetY18";
  assign OffsetTableY[18] to "{P}{A}:OffsetY19";
  assign OffsetTableY[19] to "{P}{A}:OffsetY20";
  monitor OffsetTableY;
  evflag OffsetTableYChanged;
  sync OffsetTableY OffsetTableYChanged;
long IndexOfOffsetTableY[OFFSET_TABLE_SIZE];

short OffsetTableDeleteRequest[OFFSET_TABLE_SIZE];
  assign OffsetTableDeleteRequest[0]  to "{P}{A}:Delete1";
  assign OffsetTableDeleteRequest[1]  to "{P}{A}:Delete2";
  assign OffsetTableDeleteRequest[2]  to "{P}{A}:Delete3";
  assign OffsetTableDeleteRequest[3]  to "{P}{A}:Delete4";
  assign OffsetTableDeleteRequest[4]  to "{P}{A}:Delete5";
  assign OffsetTableDeleteRequest[5]  to "{P}{A}:Delete6";
  assign OffsetTableDeleteRequest[6]  to "{P}{A}:Delete7";
  assign OffsetTableDeleteRequest[7]  to "{P}{A}:Delete8";
  assign OffsetTableDeleteRequest[8]  to "{P}{A}:Delete9";
  assign OffsetTableDeleteRequest[9]  to "{P}{A}:Delete10";
  assign OffsetTableDeleteRequest[10] to "{P}{A}:Delete11";
  assign OffsetTableDeleteRequest[11] to "{P}{A}:Delete12";
  assign OffsetTableDeleteRequest[12] to "{P}{A}:Delete13";
  assign OffsetTableDeleteRequest[13] to "{P}{A}:Delete14";
  assign OffsetTableDeleteRequest[14] to "{P}{A}:Delete15";
  assign OffsetTableDeleteRequest[15] to "{P}{A}:Delete16";
  assign OffsetTableDeleteRequest[16] to "{P}{A}:Delete17";
  assign OffsetTableDeleteRequest[17] to "{P}{A}:Delete18";
  assign OffsetTableDeleteRequest[18] to "{P}{A}:Delete19";
  assign OffsetTableDeleteRequest[19] to "{P}{A}:Delete20";
  monitor OffsetTableDeleteRequest;
  evflag OffsetTableDeleteRequestChanged;
  sync OffsetTableDeleteRequest OffsetTableDeleteRequestChanged;
long IndexOfOffsetTableDeleteRequest[OFFSET_TABLE_SIZE];

short OffsetTableDeleteAllRequest;
  assign OffsetTableDeleteAllRequest to "{P}{A}:DeleteAll";
  monitor OffsetTableDeleteAllRequest;

double OffsetTableNewEnergy;
  assign OffsetTableNewEnergy to "{P}{A}:EnergyNew";

double OffsetTableNewX;
  assign OffsetTableNewX to "{P}{A}:OffsetXNew";

double OffsetTableNewY;
  assign OffsetTableNewY to "{P}{A}:OffsetYNew";

short OffsetTableAddRequest;
  assign OffsetTableAddRequest to "{P}{A}:AddNew";
  monitor OffsetTableAddRequest;


int Index;
short AtLeastOneEntryToDelete;
int NextOffsetTableLifeNum;
short OffsetTableShouldDeleteEntry[OFFSET_TABLE_SIZE];
int Snl_SYNC;
int Status;

char *MacroP;
char *MacroA;


%{
  static void addOffsetTableEntry(SS_ID ssId, struct UserVar *pVar,
      double energy, double x, double y);
  static int compareOffsetTableEntries(SS_ID ssId, struct UserVar *pVar,
      int index1, int index2);
  static void deleteOffsetTableEntry(SS_ID ssId, struct UserVar *pVar,
      int index);
  static void deleteOffsetTableEntries(SS_ID ssId, struct UserVar *pVar,
      short shouldDelete[]);
  static int findOffsetTableEmptyEntry(SS_ID ssId, struct UserVar *pVar);
  static int findOffsetTableEntryMatchingEnergy(SS_ID ssId,
      struct UserVar *pVar, double energy);
  static int findOffsetTableOldestEntry(SS_ID ssId, struct UserVar *pVar);
  static void pvPutOffset(SS_ID ssId, struct UserVar *pVar);
  static void pvPutOffsetTable(SS_ID ssId, struct UserVar *pVar);
  static void sortOffsetTable(SS_ID ssId, struct UserVar *pVar);
  static int stepNextOffsetTableLifeNum(SS_ID ssId, struct UserVar *pVar);
  static void swapOffsetTableEntries(SS_ID ssId, struct UserVar *pVar,
      int index1, int index2);
  static void updateOffset(SS_ID ssId, struct UserVar *pVar);
}%


entry {
  Snl_SYNC = SYNC;

  IndexOfOffsetX = pvIndex(OffsetX);
  IndexOfOffsetY = pvIndex(OffsetY);

  for (Index = 0; Index < OFFSET_TABLE_SIZE; Index++) {
    IndexOfOffsetTableLifeNum[Index] = pvIndex(OffsetTableLifeNum[Index]);
    IndexOfOffsetTableSelected[Index] = pvIndex(OffsetTableSelected[Index]);
    IndexOfOffsetTableEnergy[Index] = pvIndex(OffsetTableEnergy[Index]);
    IndexOfOffsetTableX[Index] = pvIndex(OffsetTableX[Index]);
    IndexOfOffsetTableY[Index] = pvIndex(OffsetTableY[Index]);
    IndexOfOffsetTableDeleteRequest[Index] = pvIndex(
      OffsetTableDeleteRequest[Index]);
  }
}


ss Main {
  state initialize {
    when (pvConnectCount() == pvChannelCount()) {
      MacroP = macValueGet("P");
      if (!MacroP) MacroP = "";
      MacroA = macValueGet("A");
      if (!MacroA) MacroA = "";

      Status = pvGet(OffsetX);
      %%assert(pVar->Status == pvStatOK);
      Status = pvGet(OffsetY);
      %%assert(pVar->Status == pvStatOK);

      for (Index = 0; Index < OFFSET_TABLE_SIZE; Index++) {
        Status = pvGet(OffsetTableLifeNum[Index]);
        %%assert(pVar->Status == pvStatOK);
        Status = pvGet(OffsetTableSelected[Index]);
        %%assert(pVar->Status == pvStatOK);
        Status = pvGet(OffsetTableEnergy[Index]);
        %%assert(pVar->Status == pvStatOK);
      }

      NextOffsetTableLifeNum = -1;
      for (Index = 0; Index < OFFSET_TABLE_SIZE; Index++)
        if (OffsetTableLifeNum[Index] > NextOffsetTableLifeNum)
          NextOffsetTableLifeNum = OffsetTableLifeNum[Index];
      NextOffsetTableLifeNum++;

      efClear(EnergyBandChanged);
      efClear(OffsetTableDeleteRequestChanged);
      efClear(OffsetTableXChanged);
      efClear(OffsetTableYChanged);
    } state waitForSomethingToDo
  }

  state waitForSomethingToDo {
    when (efTestAndClear(EnergyChanged)) {
    } state handleEnergyChanged
    when (efTestAndClear(EnergyBandChanged)) {
    } state handleEnergyBandChanged
    when (OffsetTableDeleteAllRequest) {
    } state handleOffsetTableDeleteAllRequest
    when (efTestAndClear(OffsetTableDeleteRequestChanged)) {
    } state handleOffsetTableDeleteRequest
    when (OffsetTableAddRequest) {
    } state handleOffsetTableAddRequest
    when (efTestAndClear(OffsetTableXChanged)) {
    } state handleOffsetTableOffsetChanged
    when (efTestAndClear(OffsetTableYChanged)) {
    } state handleOffsetTableOffsetChanged
  }

  state handleEnergyChanged {
    when (delay(ENERGY_CHANGED_RATE_LIMITING_DELAY)) {
      %%updateOffset(ssId, pVar);
      %%pvPutOffset(ssId, pVar);
    } state waitForSomethingToDo
  }

  state handleEnergyBandChanged {
    when () {
      %%updateOffset(ssId, pVar);
      %%pvPutOffset(ssId, pVar);
    } state waitForSomethingToDo
  }

  state handleOffsetTableDeleteRequest {
    when () {
      AtLeastOneEntryToDelete = 0;
      for (Index = 0; Index < OFFSET_TABLE_SIZE; Index++) {
        OffsetTableShouldDeleteEntry[Index] = OffsetTableDeleteRequest[Index];
        if (OffsetTableDeleteRequest[Index]) AtLeastOneEntryToDelete = 1;
      }
      if (AtLeastOneEntryToDelete) {
        %{
          deleteOffsetTableEntries(ssId, pVar,
              pVar->OffsetTableShouldDeleteEntry);
        }%
        %%sortOffsetTable(ssId, pVar);
        %%updateOffset(ssId, pVar);
        %%pvPutOffset(ssId, pVar);
        %%pvPutOffsetTable(ssId, pVar);
        efClear(OffsetTableXChanged);
        efClear(OffsetTableYChanged);
        for (Index = 0; Index < OFFSET_TABLE_SIZE; Index++) {
          if (OffsetTableShouldDeleteEntry[Index]) {
            OffsetTableDeleteRequest[Index] = 0;
            Status = pvPut(OffsetTableDeleteRequest[Index]);
            %%assert(pVar->Status == pvStatOK);
          }
        }
      }
    } state waitForSomethingToDo
  }

  state handleOffsetTableDeleteAllRequest {
    when () {
      for (Index = 0; Index < OFFSET_TABLE_SIZE; Index++)
        OffsetTableShouldDeleteEntry[Index] = 1;
      %{
        deleteOffsetTableEntries(ssId, pVar,
            pVar->OffsetTableShouldDeleteEntry);
      }%
      %%updateOffset(ssId, pVar);
      %%pvPutOffset(ssId, pVar);
      %%pvPutOffsetTable(ssId, pVar);
      efClear(OffsetTableXChanged);
      efClear(OffsetTableYChanged);
      OffsetTableDeleteAllRequest = 0;
      Status = pvPut(OffsetTableDeleteAllRequest);
      %%assert(pVar->Status == pvStatOK);
    } state waitForSomethingToDo
  }

  state handleOffsetTableAddRequest {
    when () {
      Status = pvGet(OffsetTableNewEnergy);
      %%assert(pVar->Status == pvStatOK);
      Status = pvGet(OffsetTableNewX);
      %%assert(pVar->Status == pvStatOK);
      Status = pvGet(OffsetTableNewY);
      %%assert(pVar->Status == pvStatOK);
      %{
        addOffsetTableEntry(ssId, pVar, pVar->OffsetTableNewEnergy,
            pVar->OffsetTableNewX, pVar->OffsetTableNewY);
      }%
      %%sortOffsetTable(ssId, pVar);
      %%updateOffset(ssId, pVar);
      %%pvPutOffset(ssId, pVar);
      %%pvPutOffsetTable(ssId, pVar);
      efClear(OffsetTableXChanged);
      efClear(OffsetTableYChanged);
      OffsetTableAddRequest = 0;
      Status = pvPut(OffsetTableAddRequest);
      %%assert(pVar->Status == pvStatOK);
    } state waitForSomethingToDo
  }

  state handleOffsetTableOffsetChanged {
    when () {
      %%updateOffset(ssId, pVar);
      %%pvPutOffset(ssId, pVar);
    } state waitForSomethingToDo
  }
}


%{
static void addOffsetTableEntry(SS_ID ssId, struct UserVar *pVar,
    double energy, double x, double y) {
  int index;

  index = findOffsetTableEntryMatchingEnergy(ssId, pVar, energy);
  if (index == -1) index = findOffsetTableEmptyEntry(ssId, pVar);
  if (index == -1) {
    index = findOffsetTableOldestEntry(ssId, pVar);
    deleteOffsetTableEntry(ssId, pVar, index);
  }

  pVar->OffsetTableLifeNum[index] = stepNextOffsetTableLifeNum(ssId, pVar);
  pVar->OffsetTableSelected[index] = 0;
  pVar->OffsetTableEnergy[index] = energy;
  pVar->OffsetTableX[index] = x;
  pVar->OffsetTableY[index] = y;
}

static int compareOffsetTableEntries(SS_ID ssId, struct UserVar *pVar,
    int index1, int index2) {
  if (pVar->OffsetTableLifeNum[index1] == pVar->OffsetTableLifeNum[index2])
    return 0;
  if (pVar->OffsetTableLifeNum[index1] == OFFSET_TABLE_NIL_LIFE_NUM)
    return 1;
  if (pVar->OffsetTableLifeNum[index2] == OFFSET_TABLE_NIL_LIFE_NUM)
    return -1;
  if (pVar->OffsetTableEnergy[index1] == pVar->OffsetTableEnergy[index2])
    return 0;
  if (pVar->OffsetTableEnergy[index1] > pVar->OffsetTableEnergy[index2])
    return 1;
  else
    return -1;
}

static void deleteOffsetTableEntry(SS_ID ssId, struct UserVar *pVar,
    int index) {
  pVar->OffsetTableLifeNum[index] = OFFSET_TABLE_NIL_LIFE_NUM;
  if (pVar->OffsetTableSelected[index]) {
    pVar->OffsetTableSelected[index] = 0;
    pVar->OffsetX = 0.0;
    pVar->OffsetY = 0.0;
  }
  pVar->OffsetTableEnergy[index] = 0.0;
  pVar->OffsetTableX[index] = 0.0;
  pVar->OffsetTableY[index] = 0.0;
}

static void deleteOffsetTableEntries(SS_ID ssId, struct UserVar *pVar,
    short shouldDelete[]) {
  int i;

  for (i = 0; i < OFFSET_TABLE_SIZE; i++)
    if (shouldDelete[i])
      deleteOffsetTableEntry(ssId, pVar, i);
}

static int findOffsetTableEmptyEntry(SS_ID ssId, struct UserVar *pVar) {
  int i;

  for (i = 0; i < OFFSET_TABLE_SIZE; i++)
    if (pVar->OffsetTableLifeNum[i] == OFFSET_TABLE_NIL_LIFE_NUM) return i;
  return -1;
}

static int findOffsetTableEntryMatchingEnergy(SS_ID ssId,
    struct UserVar *pVar, double energy) {
  int i;

  for (i = 0; i < OFFSET_TABLE_SIZE; i++) {
    if (pVar->OffsetTableLifeNum[i] == OFFSET_TABLE_NIL_LIFE_NUM) continue;
    if (fabs(pVar->OffsetTableEnergy[i] - energy) <= pVar->EnergyBand)
      return i;
  }
  return -1;
}

static int findOffsetTableOldestEntry(SS_ID ssId, struct UserVar *pVar) {
  int indexOfOldestLifeNum;
  int i;

  indexOfOldestLifeNum = 0;
  for (i = 1; i < OFFSET_TABLE_SIZE; i++)
    if (pVar->OffsetTableLifeNum[i] <
        pVar->OffsetTableLifeNum[indexOfOldestLifeNum])
      indexOfOldestLifeNum = i;
  return indexOfOldestLifeNum;
}

static void pvPutOffset(SS_ID ssId, struct UserVar *pVar) {
  int i;
  pvStat status;

  for (i = 0; i < OFFSET_TABLE_SIZE; i++) {
    status = seq_pvPut(ssId, pVar->IndexOfOffsetTableSelected[i],
      pVar->Snl_SYNC);
    assert(status == pvStatOK);
  }
  status = seq_pvPut(ssId, pVar->IndexOfOffsetX, pVar->Snl_SYNC);
  assert(status == pvStatOK);
  status = seq_pvPut(ssId, pVar->IndexOfOffsetY, pVar->Snl_SYNC);
  assert(status == pvStatOK);
}

static void pvPutOffsetTable(SS_ID ssId, struct UserVar *pVar) {
  int i;
  pvStat status;

  for (i = 0; i < OFFSET_TABLE_SIZE; i++) {
    status = seq_pvPut(ssId, pVar->IndexOfOffsetTableLifeNum[i],
      pVar->Snl_SYNC);
    assert(status == pvStatOK);
    status = seq_pvPut(ssId, pVar->IndexOfOffsetTableSelected[i],
      pVar->Snl_SYNC);
    assert(status == pvStatOK);
    status = seq_pvPut(ssId, pVar->IndexOfOffsetTableEnergy[i],
      pVar->Snl_SYNC);
    assert(status == pvStatOK);
    status = seq_pvPut(ssId, pVar->IndexOfOffsetTableX[i],
      pVar->Snl_SYNC);
    assert(status == pvStatOK);
    status = seq_pvPut(ssId, pVar->IndexOfOffsetTableY[i],
      pVar->Snl_SYNC);
    assert(status == pvStatOK);
  }
}

static void sortOffsetTable(SS_ID ssId, struct UserVar *pVar) {
  int i, j;

  for (i = 0; i < OFFSET_TABLE_SIZE; i++)
    for (j = i + 1; j < OFFSET_TABLE_SIZE; j++)
      if (compareOffsetTableEntries(ssId, pVar, i, j) > 0)
        swapOffsetTableEntries(ssId, pVar, i, j);
}

static int stepNextOffsetTableLifeNum(SS_ID ssId, struct UserVar *pVar) {
  int next;

  next = pVar->NextOffsetTableLifeNum;
  pVar->NextOffsetTableLifeNum++;
  return next;
}

static void swapOffsetTableEntries(SS_ID ssId, struct UserVar *pVar,
    int index1, int index2) {
  int lifeNum;
  short selected;
  double energy;
  double x;
  double y;

  lifeNum = pVar->OffsetTableLifeNum[index1];
  selected = pVar->OffsetTableSelected[index1];
  energy = pVar->OffsetTableEnergy[index1];
  x = pVar->OffsetTableX[index1];
  y = pVar->OffsetTableY[index1];

  pVar->OffsetTableLifeNum[index1] = pVar->OffsetTableLifeNum[index2];
  pVar->OffsetTableSelected[index1] = pVar->OffsetTableSelected[index2];
  pVar->OffsetTableEnergy[index1] = pVar->OffsetTableEnergy[index2];
  pVar->OffsetTableX[index1] = pVar->OffsetTableX[index2];
  pVar->OffsetTableY[index1] = pVar->OffsetTableY[index2];

  pVar->OffsetTableLifeNum[index2] = lifeNum;
  pVar->OffsetTableSelected[index2] = selected;
  pVar->OffsetTableEnergy[index2] = energy;
  pVar->OffsetTableX[index2] = x;
  pVar->OffsetTableY[index2] = y;
}

static void updateOffset(SS_ID ssId, struct UserVar *pVar) {
  int index;
  int i;

  index = findOffsetTableEntryMatchingEnergy(ssId, pVar, pVar->Energy);
  if (index == -1) return;

  pVar->OffsetX = pVar->OffsetTableX[index];
  pVar->OffsetY = pVar->OffsetTableY[index];
  for (i = 0; i < OFFSET_TABLE_SIZE; i++)
    pVar->OffsetTableSelected[i] = i == index ? 1 : 0;
}
}%
